
-- (c) FMSE (Formal Methods in Software Engineering), MSc in Computer Science,
-- University of Minho, 2008-10
-- URL: http://www.di.uminho.pt/ensino/mestrados/mfes
--
-- Description: Checking the weakest precondition for list 'cons' to ensure th
-- no-duplicates invariant.
-- Illustrates the relational pointfree style in Alloy.
--

module NoDuplicates

open RelCalc

-- (1) Natural numbers: the minimum

sig Nat {
    succ: one Nat,
}

one sig One in Nat {}

-- succ is an injection (required):
fact{
--	RelCalc/Simple[succ, Nat]
--	RelCalc/Entire[succ, Nat]
	RelCalc/Injective[succ, Nat]
}

-- getting closer to Nat -- succ parallel to id, just for instance showing
fact {
      succ & id[Nat] = none->none
--	all n : Nat | n !in n.succ  
--	all n : Nat | n !in n.^succ
}

-- (2) Sequences: the minimum (could use Alloy standard module)

sig Data {}

sig List {
	map : Nat -> lone Data
}
--	fact{
--		all l: List | RelCalc/Simple[l.map, Data]
--	}

pred noDuplicates[l: List]{
	RelCalc/Injective[l.map, Nat]
}

pred cons[l': List, c: Data, l: List] {
	l'.map = ~(Nat->One).(Nat->c) + ~succ.(l.map)		
--	pointwise: l'.map[One]=c && all i: Nat, x : Data | x = l.map[i] => x = l'.map[i.succ]
}

pred pre[l: List, c: Data] {
	l.map . ~(Nat->c) in succ.~(Nat->One)
-- exercise: add variables and check the meaning of this precondition
}

-- as example of a function (cf. homonym Haskell function)
fun findIndices[s:set Data,l:List]: set Nat {
--	dom[l.map :> s]
       {i: Nat | some x:s | x in i.(l.map)}
}

assert FT {  -- this is the free theorem of findIndices
    all l,l':List, p: set Data, r: Data -> one Data |
          l'.map = l.map.r => findIndices[p,l'] = findIndices[r.p,l]
}

assert invPObad {
--	precondition is missing
	all c: Data, l',l:List {
		cons[l',c,l] && noDuplicates[l] => noDuplicates[l']
	}
}

assert invPO {
-- precondition is sufficient
	all c: Data, l',l:List {
		(cons[l',c,l] && noDuplicates[l] && pre[l,c]) => noDuplicates[l']
	}
}
assert invPOwp {
-- precondition is necessary
	all c: Data, l',l:List {
		cons[l',c,l] => ((noDuplicates[l] && pre[l,c]) <=> noDuplicates[l'])
	}
}

check FT 
check invPObad for 6 but 2 List
check invPOwp  for 6 but 2 List
check invPO      for 6 but 2 List

pred show () {} 
run show for 3 but 1 List 

